/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2004 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "StdAfx.h"
#include "Label.h"
#include "LabelSettings.h"
#include "xFont.h"
#include "SystemInfo.h"

#include "VerInfo.h"

//DropShadow under XP
#ifndef CS_DROPSHADOW
	#define CS_DROPSHADOW 0x00020000
#endif

//TextEditBox
#define LM_TEXTEDITBOX 9650

//PlaySound
#define   NOCOMPMAN     //  - dont include COMPMAN
#define   NODRAWDIB   //    - dont include DRAWDIB
#define   NOVIDEO     //    - dont include video capture interface
#define   NOAVIFMT    //    - dont include AVI file format structs
#define   NOMMREG     //    - dont include MMREG
#define   NOAVIFILE   //    - dont include AVIFile interface
//#define   NOMCIWND    //    - dont include AVIWnd class.
#define   NOAVICAP    //    - dont include AVICap class.
#define   NOMSACM     //    - dont include ACM stuff.
#include "vfw.h"

// Bang commands
void BangAlwaysOnTop(HWND hwndCaller, LPCSTR pszArgs);
void BangClipboardCopy(HWND hwndCaller, LPCSTR pszArgs);
void BangClipboardPaste(HWND hwndCaller, LPCSTR pszArgs);
void BangCreate(HWND hwndCaller, LPCSTR pszArgs);
void BangDestroy(HWND hwndCaller, LPCSTR pszArgs);
void BangDock(HWND hwndCaller, LPCSTR pszArgs);
void BangHide(HWND hwndCaller, LPCSTR pszArgs);
void BangLSBoxHook(HWND hwndCaller, LPCSTR pszArgs);
void BangMove(HWND hwndCaller, LPCSTR pszArgs);
void BangMoveBy(HWND hwndCaller, LPCSTR pszArgs);
void BangNext(HWND hwndCaller, LPCSTR pszArgs);
void BangPrevious(HWND hwndCaller, LPCSTR pszArgs);
void BangRefresh(HWND hwndCaller, LPCSTR pszArgs);
void BangReposition(HWND hwndCaller, LPCSTR pszArgs);
void BangRepositionBy(HWND hwndCaller, LPCSTR pszArgs);
void BangResize(HWND hwndCaller, LPCSTR pszArgs);
void BangResizeBy(HWND hwndCaller, LPCSTR pszArgs);
void BangScroll(HWND hwndCaller, LPCSTR pszArgs);
void BangSetAlpha(HWND hwndCaller, LPCSTR pszArgs);
void BangSetAnimation(HWND hwndCaller, LPCSTR pszArgs);
void BangSetFontColor(HWND hwndCaller, LPCSTR pszArgs);
void BangSetFontShadowColor(HWND hwndCaller, LPCSTR pszArgs);
void BangSetText(HWND hwndCaller, LPCSTR pszArgs);
void BangShow(HWND hwndCaller, LPCSTR pszArgs);
void BangShowHide(HWND hwndCaller, LPCSTR pszArgs);
void BangToggle(HWND hwndCaller, LPCSTR pszArgs);
void BangGhosted(HWND hwndCaller, LPCSTR pszArgs);
void BangUpdate(HWND hwndCaller, LPCSTR pszArgs);

void BangInfoExport(HWND hwndCaller, LPCSTR pszArgs);

void BangParseEvars(HWND hwndCaller, LPCSTR pszArgs);
void BangSetEvars(HWND hwndCaller, LPCSTR pszArgs);

void BangPlaySound(HWND hwndCaller, LPCSTR pszArgs);

HWND m_hWndSoundFile = NULL;

HINSTANCE hInstance;
HWND messageHandler;

#ifndef SM_CXVIRTUALSCREEN
	#define SM_CXVIRTUALSCREEN      78
#endif
#ifndef SM_CYVIRTUALSCREEN
	#define SM_CYVIRTUALSCREEN      79
#endif

int screenX = GetSystemMetrics(SM_CXVIRTUALSCREEN);
int	screenY = GetSystemMetrics(SM_CYVIRTUALSCREEN);

int bitsperpixel = atoi( QueryPixelDepth().c_str() );

LabelList labelList;

int lsMessages[] = {
	LM_GETREVID,
	LM_REFRESH,
	0
};

Label *lookupLabel( const string &name );

#define LM_UPDATEBG (WM_USER + 1)

LRESULT WINAPI MessageHandlerProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
		case LM_GETREVID:
		{
			UINT uLength;
			StringCchPrintf((char*)lParam, 64, "%s.dll: %s", V_NAME, V_VERSION);
			
			if (SUCCEEDED(StringCchLength((char*)lParam, 64, &uLength)))
				return uLength;

			lParam = NULL;
			return 0;
		}
		
		case LM_REFRESH:
		{
			StringList labelNames = GetxRCNameList("*label");

			// refresh the "AllLabels" configuration
			for(LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if(!(*lit)->getBox())
				{
					// destroy all labels that no longer exist and that are not in a box
					for(StringListIterator sit = labelNames.begin(); sit != labelNames.end(); sit++)
					{
						if(stricmp((*sit).c_str(), (*lit)->getName().c_str()) == NULL)
							break;
					}
					if (sit == labelNames.end())
					{
						(*lit)->hide();
						labelList.remove(*lit);
						delete *lit;
						continue;
					}
				}

				// we can reconfigure all other labels, even if they are "boxed"
				(*lit)->reconfigure(true);
				(*lit)->repaint(true);
			}

			// create the rest
			for(StringListIterator sit = labelNames.begin(); sit != labelNames.end(); sit++)
			{
				Label *label = lookupLabel(*sit);
				
				if (!label) 
				{
					label = new Label(*sit);
					labelList.insert(labelList.end(), label);
					
					if ( IsOS(OS_XP) && GetxRCBool((*sit).c_str(), "DropShadow") )
						label->load(hInstance, 0, true);
					else 
						label->load(hInstance);

					//ModuleHook 
					if (!label->moduleHook.empty())
					{
						for(StringListIterator sit = label->moduleHook.begin(); sit != label->moduleHook.end(); sit++)
						{
							string hWndbuffer = IntToString( (int) label->gethWnd() );
							
							string config = (*sit);
							config.append( " \"$themedir$theme.rc\" " );
							config.append( hWndbuffer );

							LSExecute(NULL, config.c_str(), SW_SHOWNORMAL);
						}

						label->hasHookedModules = true;
						label->moduleHook.clear();
					}

					if (!label->isStartHidden())
						label->show();
					else
						label->hide();

					//Hooked Modules cannot Dock!
					if ( label->getBox() != NULL )
						continue;

					char name[64], position[64];
					char *buffers[] = {name, position};
					for(StringListIterator sit = label->dockedToList.begin(); sit != label->dockedToList.end(); sit++)
					{
						LCTokenize((*sit).c_str(), buffers, 2, NULL);
						if ( stricmp(position, "left") == NULL )
							label->dockedtoSide = 0;
						else if ( stricmp(position, "top") == NULL )
							label->dockedtoSide = 1;
						else if ( stricmp(position, "right") == NULL )
							label->dockedtoSide = 2;
						else if ( stricmp(position, "bottom") == NULL )
							label->dockedtoSide = 3;
						else if ( stricmp(position, "topleft") == NULL )
							label->dockedtoSide = 4;
						else if ( stricmp(position, "topright") == NULL )
							label->dockedtoSide = 5;
						else if ( stricmp(position, "bottomleft") == NULL )
							label->dockedtoSide = 6;
						else if ( stricmp(position, "bottomright") == NULL )
							label->dockedtoSide = 7;

						for(LabelListIterator plit = labelList.begin(); plit != labelList.end(); plit++)
						{
							if( stricmp(name, (*plit)->getName().c_str()) == NULL )
							{
								(*plit)->hasDockedLabelList.insert((*plit)->hasDockedLabelList.end(), label);
								break;
							}
						}
					}
				}
			}

			return 0;
		}

		case LM_UPDATEBG:
		{
			PaintDesktopEx(0, 0, 0, 0, 0, 0, 0, TRUE);

			for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if ( (*lit)->getBox() == 0 && (*lit)->getTransparencyMode() == 2 )
					(*lit)->repaint(true);
			}

			return 0;
		}

		case WM_DISPLAYCHANGE:
		case WM_SETTINGCHANGE:
		{
			PostMessage(hWnd, LM_UPDATEBG, 0, 0);
			return 0;
		}

		case WM_TIMER:
		{
			if (m_hWndSoundFile != NULL)
			{
				if (MCIWndGetEnd(m_hWndSoundFile) == MCIWndGetPosition(m_hWndSoundFile))
				{
					MCIWndStop(m_hWndSoundFile);
					MCIWndDestroy(m_hWndSoundFile);
					m_hWndSoundFile = NULL;

					KillTimer(messageHandler, NULL);
				}
			}
			else
			{
				KillTimer(messageHandler, NULL);
			}
			break;
		}
	}

	return DefWindowProc(hWnd, message, wParam, lParam);
}

BOOL __stdcall _DllMainCRTStartup(HINSTANCE hInst, DWORD fdwReason, LPVOID lpvRes);

BOOL __stdcall _DllMainCRTStartup(HINSTANCE hInst, DWORD fdwReason, LPVOID lpvRes)
{
	// We don't need thread notifications for what we're doing.  Thus, get
	// rid of them, thereby eliminating some of the overhead of this DLL
	DisableThreadLibraryCalls(hInst);

	return TRUE;
}

int initModuleEx(HWND hParent, HINSTANCE hInstance, const char *lsPath)
{
	WNDCLASSEX wc;

	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = CS_GLOBALCLASS | CS_DBLCLKS ;
	wc.lpfnWndProc = Label::windowProcedure;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = sizeof(Label *);
	wc.hInstance = hInstance;
	wc.hbrBackground = 0;
	wc.hCursor = LoadCursor(0, IDC_ARROW);
	wc.hIcon = 0;
	wc.lpszMenuName = 0;
	wc.lpszClassName = "xLabel";
	wc.hIconSm = 0;

	RegisterClassEx(&wc);

	if(IsOS(OS_XP))
	{
		wc.cbSize = sizeof(WNDCLASSEX);
		wc.style = CS_GLOBALCLASS | CS_DBLCLKS | CS_DROPSHADOW;
		wc.lpfnWndProc = Label::windowProcedure;
		wc.cbClsExtra = 0;
		wc.cbWndExtra = sizeof(Label *);
		wc.hInstance = hInstance;
		wc.hbrBackground = 0;
		wc.hCursor = LoadCursor(0, IDC_ARROW);
		wc.hIcon = 0;
		wc.lpszMenuName = 0;
		wc.lpszClassName = "xLabelXP";
		wc.hIconSm = 0;

		RegisterClassEx(&wc);
	}

	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = CS_GLOBALCLASS;
	wc.lpfnWndProc = MessageHandlerProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hbrBackground = 0;
	wc.hCursor = LoadCursor(0, IDC_ARROW);
	wc.hIcon = 0;
	wc.lpszMenuName = 0;
	wc.lpszClassName = "xLabelMessageHandler";
	wc.hIconSm = 0;

	RegisterClassEx(&wc);

	messageHandler = CreateWindowEx(WS_EX_TOOLWINDOW,
		"xLabelMessageHandler",
		0,
		WS_POPUP,
		0, 0, 0, 0, 
		0,
		0,
		hInstance,
		0);

	if (!messageHandler)
		return 1;
	
	SendMessage(GetLitestepWnd(),
		LM_REGISTERMESSAGE,
		(WPARAM) messageHandler,
		(LPARAM) lsMessages);

	::hInstance = hInstance;

	defaultSettings = new LabelSettings();
	systemInfo = new SystemInfo();

	AddBangCommand("!LabelAlwaysOnTop", BangAlwaysOnTop);
	AddBangCommand("!LabelClipboardCopy", BangClipboardCopy);
	AddBangCommand("!LabelClipboardPaste", BangClipboardPaste);
	AddBangCommand("!LabelCreate", BangCreate);
	AddBangCommand("!LabelDestroy", BangDestroy);
	AddBangCommand("!LabelDock", BangDock);
	AddBangCommand("!LabelHide", BangHide);
	AddBangCommand("!LabelLSBoxHook", BangLSBoxHook);
	AddBangCommand("!LabelMove", BangMove);
	AddBangCommand("!LabelMoveBy", BangMoveBy);
	AddBangCommand("!LabelNext", BangNext);
	AddBangCommand("!LabelPrevious", BangPrevious);
	AddBangCommand("!LabelRefresh", BangRefresh);
	AddBangCommand("!LabelReposition", BangReposition);
	AddBangCommand("!LabelRepositionBy", BangRepositionBy);
	AddBangCommand("!LabelResize", BangResize);
	AddBangCommand("!LabelResizeBy", BangResizeBy);
	AddBangCommand("!LabelScroll", BangScroll);
	if(IsOS(OS_2KXP))
		AddBangCommand("!LabelSetAlpha", BangSetAlpha);
	AddBangCommand("!LabelSetAnimation", BangSetAnimation);
	AddBangCommand("!LabelSetFontColor", BangSetFontColor);
	AddBangCommand("!LabelSetFontShadowColor", BangSetFontShadowColor);
	AddBangCommand("!LabelSetText", BangSetText);
	AddBangCommand("!LabelShow", BangShow);
	AddBangCommand("!LabelShowHide", BangShowHide);
	AddBangCommand("!LabelToggle", BangToggle);
	AddBangCommand("!LabelGhosted", BangGhosted);
	AddBangCommand("!LabelUpdate", BangUpdate);

	AddBangCommand("!LabelInfoExport", BangInfoExport);

	AddBangCommand("!ParseEvars", BangParseEvars);
	AddBangCommand("!SetEvar", BangSetEvars);

	AddBangCommand("!PlaySound", BangPlaySound);

	StringList labelNames = GetxRCNameList("*label");

	for (StringListIterator sit = labelNames.begin(); sit != labelNames.end(); sit++)
	{
		if ( lookupLabel((*sit)) != NULL )
			continue;

		Label *label = new Label(*sit);
		labelList.insert(labelList.end(), label);
		
		if ( IsOS(OS_XP) && GetxRCBool((*sit).c_str(), "DropShadow") )
			label->load(hInstance, 0, true);
		else 
			label->load(hInstance);
		
		//ModuleHook
		if (!label->moduleHook.empty())
		{
			for(StringListIterator sit = label->moduleHook.begin(); sit != label->moduleHook.end(); sit++)
			{
				string hWndbuffer = IntToString( (int) label->gethWnd() );
				
				string config = (*sit);
				config.append( " \"$themedir$theme.rc\" " );
				config.append( hWndbuffer );

				LSExecute(NULL, config.c_str(), SW_SHOWNORMAL);
			}

			label->hasHookedModules = true;
			label->moduleHook.clear();
		}
		
		//Initial Show
		if (!label->isStartHidden())
			label->show();
		else
			label->hide();
	}

	//Docking Setup
	for(LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
	{
		//Hooked Modules cannot Dock!
		if ( (*lit)->getBox() != NULL )
			continue;

		char name[64], position[64];
		char *buffers[] = {name, position};
		for(StringListIterator sit = (*lit)->dockedToList.begin(); sit != (*lit)->dockedToList.end(); sit++)
		{
			LCTokenize((*sit).c_str(), buffers, 2, NULL);
			if ( stricmp(position, "left") == NULL )
				(*lit)->dockedtoSide = 0;
			else if ( stricmp(position, "top") == NULL )
				(*lit)->dockedtoSide = 1;
			else if ( stricmp(position, "right") == NULL )
				(*lit)->dockedtoSide = 2;
			else if ( stricmp(position, "bottom") == NULL )
				(*lit)->dockedtoSide = 3;
			else if ( stricmp(position, "topleft") == NULL )
				(*lit)->dockedtoSide = 4;
			else if ( stricmp(position, "topright") == NULL )
				(*lit)->dockedtoSide = 5;
			else if ( stricmp(position, "bottomleft") == NULL )
				(*lit)->dockedtoSide = 6;
			else if ( stricmp(position, "bottomright") == NULL )
				(*lit)->dockedtoSide = 7;

			for(LabelListIterator plit = labelList.begin(); plit != labelList.end(); plit++)
			{
				if( stricmp( name, (*plit)->getName().c_str()) == NULL )
				{
					(*plit)->hasDockedLabelList.insert((*plit)->hasDockedLabelList.end(), (*lit));
					break;
				}
			}
		}
	}

	return 0;
}

extern HDC hdcDesktop;
extern HBITMAP hbmDesktop;

void quitModule(HINSTANCE hInstance)
{
	if (m_hWndSoundFile != NULL)
	{
		MCIWndStop(m_hWndSoundFile);
		MCIWndDestroy(m_hWndSoundFile);
		m_hWndSoundFile = NULL;
	}

	RemoveBangCommand("!LabelAlwaysOnTop");
	RemoveBangCommand("!LabelClipboardCopy");
	RemoveBangCommand("!LabelClipboardPaste");
	RemoveBangCommand("!LabelCreate");
	RemoveBangCommand("!LabelDestroy");
	RemoveBangCommand("!LabelDock");
	RemoveBangCommand("!LabelHide");
	RemoveBangCommand("!LabelLSBoxHook");
	RemoveBangCommand("!LabelMove");
	RemoveBangCommand("!LabelMoveBy");
	RemoveBangCommand("!LabelNext");
	RemoveBangCommand("!LabelPrevious");
	RemoveBangCommand("!LabelRefresh");
	RemoveBangCommand("!LabelReposition");
	RemoveBangCommand("!LabelRepositionBy");
	RemoveBangCommand("!LabelResize");
	RemoveBangCommand("!LabelResizeBy");
	RemoveBangCommand("!LabelScroll");
	if(IsOS(OS_2KXP))
		RemoveBangCommand("!LabelSetAlpha");
	RemoveBangCommand("!LabelSetAnimation");
	RemoveBangCommand("!LabelSetFontColor");
	RemoveBangCommand("!LabelSetFontShadowColor");
	RemoveBangCommand("!LabelSetText");
	RemoveBangCommand("!LabelShow");
	RemoveBangCommand("!LabelShowHide");
	RemoveBangCommand("!LabelToggle");
	RemoveBangCommand("!LabelGhosted");
	RemoveBangCommand("!LabelUpdate");

	RemoveBangCommand("!LabelInfoExport");

	RemoveBangCommand("!ParseEvars");
	RemoveBangCommand("!SetEvar");

	RemoveBangCommand("!PlaySound");

	for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
	{
		(*lit)->hide();
		delete *lit;
	}

	labelList.clear();

	SendMessage(GetLitestepWnd(),
		LM_UNREGISTERMESSAGE,
		(WPARAM) messageHandler,
		(LPARAM) lsMessages);

	DestroyWindow(messageHandler);

	UnregisterClass("xLabel", hInstance);
	if(IsOS(OS_XP))
		UnregisterClass("xLabelXP", hInstance);
	UnregisterClass("xLabelMessageHandler", hInstance);

	if (defaultSettings)
		delete defaultSettings;
	if (systemInfo)
		delete systemInfo;
	
	if (hbmDesktop != 0 || hdcDesktop != 0)
	{
		hbmDesktop = (HBITMAP) SelectObject(hdcDesktop, hbmDesktop);
		DeleteObject(hbmDesktop);
		DeleteDC(hdcDesktop);
	}
}

//Text EditBox
//-----------------------------------------------------
LRESULT APIENTRY TextEditProc(HWND hText, UINT msg, WPARAM wParam, LPARAM lParam)
{
	Label* label = (Label *) GetWindowLong(GetParent(hText), 0);

	char buf[MAX_PATH] = {0};

	switch(msg)
	{
		case WM_CHAR:
		{
			if( wParam == VK_RETURN )
			{
				GetWindowText(hText, buf, sizeof(buf));

				SendMessage(label->gethWnd(), LM_TEXTEDITBOX, 0, (LONG) buf);
			}
			else if ( wParam == VK_ESCAPE )
			{
				ParseBangCommand(label->gethWnd(), "!labelhide", (label->getName() + " .editbox").c_str());
			}
		}
		break;

		case WM_SETCURSOR:

			string process = label->getTextCursorPath();
			if ( process.empty() )
			{
				break;
			}
			else if ( stricmp(process.c_str(), ".none") == NULL )
			{
				return TRUE;
			}
			else if ( label->getTextCursor() )
			{
				SetCursor( label->getTextCursor() );
				return TRUE;
			}
		
		return FALSE;
    }

	return CallWindowProc(label->wpOrigEditProc, hText, msg, wParam, lParam);
}
//-----------------------------------------------------

Label *lookupLabel( const string &name )
{
	for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
	{
		if (stricmp(name.c_str(), (*lit)->getName().c_str()) == NULL)
			return *lit;
	}

	return NULL;
}

// !LabelInfoExport < clipboard | mzscript var | $Evar$> <Text Escape Sequence>
void BangInfoExport(HWND hwndCaller, LPCSTR pszArgs)
{
	string Bang = "";
	char namevar[64], textvar[MAX_PATH_LENGTH];
	char *buffers[] = {namevar, textvar};

	if (LCTokenize(pszArgs, buffers, 2, NULL) == 2)
	{	
		Label *label = lookupLabel(namevar);

		if(label)
		{
			string evar = textvar;

			if (strstr(evar.c_str(), "%#") != NULL)
			{
				while ( strstr(evar.c_str(), "%#") != NULL )
				{
					evar.replace(evar.find("%#"), 2, "", 0);
				}

				Bang = ("\"" + label->getText() + "\"");
				PrefixedLSSetVariable( evar, "", Bang );
			}
			else
			{
				Bang = evar;
				Bang.append(" \"" + label->getText() + "\"");
				ParseBangCommand(NULL, "!varset", Bang.c_str());
			}
		}
		else
		{
			string result = "";
			string evar = namevar;
			bool dynamic = false;
			result = systemInfo->processLabelText(textvar, (*labelList.begin()), &dynamic);

			if ( stricmp(evar.c_str(), "clipboard") == NULL)
			{
				SetClipboardText(NULL, result.c_str());
			}
			else if (strstr(evar.c_str(), "%#") != NULL)
			{
				while ( strstr(evar.c_str(), "%#") != NULL )
				{
					evar.replace(evar.find("%#"), 2, "", 0);
				}

				Bang = ("\"" + result + "\"");
				PrefixedLSSetVariable( evar, "", Bang );
			}
			else
			{
				Bang = evar;
				Bang.append(" \"" + result + "\"");
				ParseBangCommand(NULL, "!varset", Bang.c_str());
			}
		}
	}
}

// !LabelSetAnimation <label> <on | off> [<loops>]
void BangSetAnimation(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], option[16], loops[16];
	char *buffers[] = {name, option, loops};

	int numTokens = LCTokenize(pszArgs, buffers, 3, NULL);
	if (numTokens >= 2)
	{
		Label *label = lookupLabel(name);

		if (label)
		{
			if (numTokens == 2)
			{
				if (stricmp(option, "off") == NULL)
					label->setAnimation(false);
				else if (stricmp(option, "on") == NULL)
					label->setAnimation(true);
			}
			else if (numTokens == 3)
			{
				if (stricmp(option, "off") == NULL)
					label->setAnimation(false);
				else if (stricmp(option, "on") == NULL)
					label->setAnimation(true, atoi(loops));
			}
		}
	}
}

// !LabelAlwaysOnTop <label> <on|true, off|false, toggle>
void BangAlwaysOnTop(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], option[16];
	char *buffers[] = {name, option};

	LCTokenize(pszArgs, buffers, 2, NULL);

	Label *label = lookupLabel(name);
	if(label)
	{
		if (stricmp(option, "false") == NULL || stricmp(option, "off") == NULL)
			label->setAlwaysOnTop(false);
		else if (stricmp(option, "toggle") == NULL || stricmp(option, "switch") == NULL)
			label->setAlwaysOnTop(!label->isAlwaysOnTop());
		else
			label->setAlwaysOnTop(true);
	}
	else
	{
		LabelListIterator lit;
		for( lit = labelList.begin(); lit != labelList.end(); lit++ )
		{
			if( stricmp((*lit)->labelGroup.c_str(), name) == NULL )
			{
				if (stricmp(option, "false") == NULL || stricmp(option, "off") == NULL)
					(*lit)->setAlwaysOnTop(false);
				else if (stricmp(option, "toggle") == NULL || stricmp(option, "switch") == NULL)
					(*lit)->setAlwaysOnTop(!(*lit)->isAlwaysOnTop());
				else
					(*lit)->setAlwaysOnTop(true);
			}
		}
	}
}

// !LabelClipboardCopy <label> <prefix>
void BangClipboardCopy(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], prefix[256];
	char *buffers[] = {name, prefix};
	int numTokens = LCTokenize(pszArgs, buffers, 2, NULL);

	if(numTokens >= 1)
	{
		Label *label = lookupLabel(name);

		if(label)
			label->clipboardCopy((numTokens >= 2) ? prefix : "");
	}
}

// !LabelClipboardPaste <label> <prefix>
void BangClipboardPaste(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], prefix[256];
	char *buffers[] = {name, prefix};
	int numTokens = LCTokenize(pszArgs, buffers, 2, NULL);

	if(numTokens >= 1)
	{
		Label *label = lookupLabel(name);

		if(label)
			label->clipboardPaste((numTokens >= 2) ? prefix : "");
	}
}

// !LabelCreate <label>
void BangCreate(HWND hwndCaller, LPCSTR pszArgs)
{
	if ( lookupLabel( pszArgs ) != NULL )
		return;

	Label *label = new Label(pszArgs);
	labelList.insert(labelList.end(), label);

	if ( IsOS(OS_XP) && GetxRCBool(pszArgs, "DropShadow") )
		label->load(hInstance, 0, true);
	else 
		label->load(hInstance);

	//ModuleHook
	if (!label->moduleHook.empty())
	{
		for(StringListIterator sit = label->moduleHook.begin(); sit != label->moduleHook.end(); sit++)
		{
			string hWndbuffer = IntToString( (int) label->gethWnd() );
			
			string config = (*sit);
			config.append( " \"$themedir$theme.rc\" " );
			config.append( hWndbuffer );

			LSExecute(NULL, config.c_str(), SW_SHOWNORMAL);
		}

		label->hasHookedModules = true;
		label->moduleHook.clear();
	}

	if (!label->isStartHidden())
		label->show();
	else
		label->hide();

	//Docking
	char name[64], position[64];
	char *buffers[] = {name, position};
	for(StringListIterator sit = label->dockedToList.begin(); sit != label->dockedToList.end(); sit++)
	{
		LCTokenize((*sit).c_str(), buffers, 2, NULL);
		if ( stricmp(position, "left") == NULL )
			label->dockedtoSide = 0;
		else if ( stricmp(position, "top") == NULL )
			label->dockedtoSide = 1;
		else if ( stricmp(position, "right") == NULL )
			label->dockedtoSide = 2;
		else if ( stricmp(position, "bottom") == NULL )
			label->dockedtoSide = 3;
		else if ( stricmp(position, "topleft") == NULL )
			label->dockedtoSide = 4;
		else if ( stricmp(position, "topright") == NULL )
			label->dockedtoSide = 5;
		else if ( stricmp(position, "bottomleft") == NULL )
			label->dockedtoSide = 6;
		else if ( stricmp(position, "bottomright") == NULL )
			label->dockedtoSide = 7;

		for(LabelListIterator plit = labelList.begin(); plit != labelList.end(); plit++)
		{
			if( stricmp( (*sit).c_str(), (*plit)->getName().c_str()) == NULL )
			{
				(*plit)->hasDockedLabelList.insert((*plit)->hasDockedLabelList.end(), label);
				break;
			}
		}
	}
}

// !LabelDestroy <label>
void BangDestroy(HWND hwndCaller, LPCSTR pszArgs)
{
	Label *label = lookupLabel(pszArgs);
	
	if (label)
	{
		label->hide();
		labelList.remove(label);
		delete label;
	}
	else
	{
		for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
		{
			if( stricmp((*lit)->labelGroup.c_str(), pszArgs) == NULL )
			{
				(*lit)->hide();
				labelList.remove((*lit));
				delete (*lit);
			}
		}
	}
}

// !LabelDock <label> <label> <"true", "false"> <modifier>
void BangDock(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], dockto[64], state[32], modifier[32];
	char *buffers[] = {name, dockto, state, modifier};

	if(LCTokenize(pszArgs, buffers, 4, NULL) >= 2)
	{	
		Label *label = lookupLabel(name);
		Label *labeldockto = lookupLabel(dockto);

		if (stricmp( state, "false" ) == NULL)
		{
			if(label && labeldockto)
			{
				for(LabelListIterator lit = labeldockto->hasDockedLabelList.begin(); lit != labeldockto->hasDockedLabelList.end(); lit++)
				{
					if ( stricmp( name, (*lit)->getName().c_str() ) == NULL )
					{
						labeldockto->hasDockedLabelList.erase ( lit );
						break;
					}
				}	
			}
		}
		else
		{
			if(label && labeldockto)
			{
				bool alreadyFound = false;
				for(LabelListIterator lit = labeldockto->hasDockedLabelList.begin(); lit != labeldockto->hasDockedLabelList.end(); lit++)
				{
					if ( stricmp( name, (*lit)->getName().c_str() ) == NULL )
					{
						alreadyFound = true;
						break;
					}
				}
				if ( !alreadyFound )
				{
					if ( stricmp(modifier, "left") == NULL )
						label->dockedtoSide = 0;
					else if ( stricmp(modifier, "top") == NULL )
						label->dockedtoSide = 1;
					else if ( stricmp(modifier, "right") == NULL )
						label->dockedtoSide = 2;
					else if ( stricmp(modifier, "bottom") == NULL )
						label->dockedtoSide = 3;
					else if ( stricmp(modifier, "topleft") == NULL )
						label->dockedtoSide = 4;
					else if ( stricmp(modifier, "topright") == NULL )
						label->dockedtoSide = 5;
					else if ( stricmp(modifier, "bottomleft") == NULL )
						label->dockedtoSide = 6;
					else if ( stricmp(modifier, "bottomright") == NULL )
						label->dockedtoSide = 7;
					else
						label->dockedtoSide = -1;

					labeldockto->hasDockedLabelList.insert ( labeldockto->hasDockedLabelList.end(), label );
				}
			}
		}
	}
}

// !LabelShow <label>
void BangShow(HWND hwndCaller, LPCSTR pszArgs)
{
	bool textEditBox = false;
	char name[64], textbox[64];
	char *buffers[] = {name, textbox};

	LCTokenize(pszArgs, buffers, 2, NULL);

	if ( stricmp(textbox, ".editbox") == NULL )
		textEditBox = true;

	Label *label = lookupLabel(name);

	if (label)
		label->show(true, textEditBox);
	else
	{
		for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
		{
			if( stricmp((*lit)->labelGroup.c_str(), name) == NULL )
				(*lit)->show(true, textEditBox);
		}
	}
}

// !LabelShowHide <label> <timeout>
void BangShowHide(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], A[64];
	char *buffers[] = {name, A};

	if (LCTokenize(pszArgs, buffers, 2, NULL) >= 2)
	{
		int timeout = atoi(A);

		Label *label = lookupLabel(name);

		if (label)
			label->showHide(timeout, true);
		else
		{
			for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if ( stricmp((*lit)->labelGroup.c_str(), name) == NULL )
					(*lit)->showHide(timeout, true);
			}
		}
	}
}

// !LabelHide <label>
void BangHide(HWND hwndCaller, LPCSTR pszArgs)
{
	bool textEditBox = false;
	char name[64], textbox[64];
	char *buffers[] = {name, textbox};

	LCTokenize(pszArgs, buffers, 2, NULL);

	if ( stricmp(textbox, ".editbox") == NULL )
		textEditBox = true;

	Label *label = lookupLabel(name);
	
	if (label)
		label->hide(true, textEditBox);
	else
	{
		for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++ )
		{
			if( stricmp((*lit)->labelGroup.c_str(), name) == NULL )
				(*lit)->hide(true, textEditBox);
		}
	}
}

// !LabelLSBoxHook <labelname>
void BangLSBoxHook(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], rest[MAX_PATH_LENGTH];
	char *buffers[] = {name};

	int tokens = LCTokenize(pszArgs, buffers, 1, rest);

	const char *handle = strrchr(pszArgs,' ');

	HWND hwndBox = (HWND)atoi(handle+1);

	Label *label = lookupLabel(name);

	if (label)
	{
		label->setBox(hwndBox);
		
		//AlphaMap Support
		char windowClass[MAX_PATH_LENGTH] = {0};
		GetClassName( hwndBox, windowClass, MAX_PATH_LENGTH );
		if (stricmp(windowClass, "xlabel") == NULL || stricmp(windowClass, "xlabelxp") == NULL)
		{
			char windowCaption[MAX_PATH_LENGTH] = {0};
					
			GetWindowText( hwndBox, windowCaption, MAX_PATH_LENGTH );
		
			for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if (stricmp(windowCaption, (*lit)->getName().c_str()) == NULL)
				{
					if ( (*lit)->getAlphaMap() )
						label->parentIsAlphaMapped = true;
					break;
				}
			}
		}
		//----------------

		label->update();		
	}
	else
	{	
		label = new Label(name);
		labelList.insert(labelList.end(), label);

		//AlphaMap Support
		char windowClass[MAX_PATH_LENGTH] = {0};
		GetClassName( hwndBox, windowClass, MAX_PATH_LENGTH );
		if (stricmp(windowClass, "xlabel") == NULL || stricmp(windowClass, "xlabelxp") == NULL)
		{
			char windowCaption[MAX_PATH_LENGTH] = {0};
					
			GetWindowText( hwndBox, windowCaption, MAX_PATH_LENGTH );
		
			for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if (stricmp(windowCaption, (*lit)->getName().c_str()) == NULL)
				{
					if ( (*lit)->getAlphaMap() )
						label->parentIsAlphaMapped = true;
					break;
				}
			}
		}
		//----------------

		if ( IsOS(OS_XP) && GetxRCBool(name, "DropShadow") )
			label->load(hInstance, hwndBox, true);
		else 
			label->load(hInstance, hwndBox);

		//ModuleHook
		if (!label->moduleHook.empty())
		{
			for(StringListIterator sit = label->moduleHook.begin(); sit != label->moduleHook.end(); sit++)
			{
				string hWndbuffer = IntToString( (int) label->gethWnd() );
				
				string config = (*sit);
				config.append( " \"$themedir$theme.rc\" " );
				config.append( hWndbuffer );

				LSExecute(NULL, config.c_str(), SW_SHOWNORMAL);
			}

			label->hasHookedModules = true;
			label->moduleHook.clear();
		}

		if (!label->isStartHidden())
			label->show();
		else
			label->hide();
	}
}

// !LabelMove <label> <x> <y> [<steps> <time>]
void BangMove(HWND hwndCaller, LPCSTR pszArgs)
{
	int x = 0;
	int y = 0;
	char name[64], A[16], B[16], C[16], D[16];
	char *buffers[] = {name, A, B, C, D};

	if (LCTokenize(pszArgs, buffers, 5, NULL) >= 3)
	{
		Label *label = lookupLabel(name);

		if (label)
		{
			if (label->getBox() == NULL)
			{
				x = xParseCoord(A, 0, screenX);
				y = xParseCoord(B, 0, screenY);
			}
			else
			{
				RECT rc;
				GetClientRect(label->getBox(), &rc);
				x = xParseCoord(A, 0, rc.right-rc.left);
				y = xParseCoord(B, 0, rc.bottom-rc.top);
			}
			int step = atoi( C );
			int time = atoi( D );

			label->reposition(1, x, y, 0, 0, step, time);
		}
	}
}

// !LabelMoveBy <label> <x> <y> [<steps> <time>] [bool s]
void BangMoveBy(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], A[16], B[16], C[16], D[16];
	char *buffers[] = {name, A, B, C, D};

	if (LCTokenize(pszArgs, buffers, 5, NULL) >= 3)
	{
		Label *label = lookupLabel(name);

		int x = atoi( A );
		int y = atoi( B );

		int step = atoi( C );
		int time = atoi( D );

		if (label)
			label->reposition(1, x, y, 0, 0, step, time, true);
		else
		{
			for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if ( stricmp((*lit)->labelGroup.c_str(), name) == NULL )
					(*lit)->reposition(1, x, y, 0, 0, step, time, true);
			}
		}
	}
}

// !LabelNext <label>
void BangNext(HWND hwndCaller, LPCSTR pszArgs)
{
	Label *label = lookupLabel(pszArgs);

	if (label)
		label->next();
}

// !LabelPrevious <label>
void BangPrevious(HWND hwndCaller, LPCSTR pszArgs)
{
	Label *label = lookupLabel(pszArgs);

	if (label)
		label->previous();
}

// !LabelRefresh <label>
void BangRefresh(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], setting[256], value[MAX_PATH_LENGTH];
	char *buffers[] = {name, setting};

	int tokens = LCTokenize(pszArgs, buffers, 2, value);

	Label *label = lookupLabel(name);

	if (label)
	{
		if (tokens == 1)
		{
			label->reconfigure(true);
			label->repaint(true);
		}

		if (tokens == 2)
		{
			string updatevalue = value;
			
			//Real Negative Fix
			if (stricmp(setting, "x") == NULL || stricmp(setting, "y") == NULL)
				updatevalue.replace(updatevalue.find("-"), 1, "~", 1);

			PrefixedLSSetVariable( (string)setting, "", updatevalue );
			label->reconfigure(true);
			label->repaint(true);
		}
	}
	else
	{
		for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
		{
			if ( stricmp((*lit)->labelGroup.c_str(), name) == NULL )
			{
				if (tokens == 1)
				{
					label->reconfigure(true);
					label->repaint(true);
				}

				if (tokens == 2)
				{
					string updatevalue = value;
					
					//Real Negative Fix
					if (stricmp(setting, "x") == NULL || stricmp(setting, "y") == NULL)
						updatevalue.replace(updatevalue.find("-"), 1, "~", 1);

					string worker = setting;

					if (strstr(setting, "*") != NULL)
						worker.replace(worker.find("*"), 1, (*lit)->getName().c_str(), (*lit)->getName().length());

					PrefixedLSSetVariable( worker, "", (string)value );
					(*lit)->reconfigure(true);
					(*lit)->repaint(true);
				}
			}
		}
	}
}

// !LabelReposition <label> <x> <y> <width> <height> [<steps>] [<time>]
void BangReposition(HWND hwndCaller, LPCSTR pszArgs)
{
	int x = 0;
	int y = 0;
	int w = 64;
	int h = 64;
	char name[64], A[16], B[16], C[16], D[16], E[16], F[16];
	char *buffers[] = {name, A, B, C, D, E, F};

	if(LCTokenize(pszArgs, buffers, 7, NULL) >= 5)
	{
		Label *label = lookupLabel(name);

		if (label)
		{
			if (label->getBox() == NULL)
			{
				x = xParseCoord(A, 0, screenX);
				y = xParseCoord(B, 0, screenY);
				w = xParseDimen(C, 0, screenX);
				h = xParseDimen(D, 0, screenY);
			}
			else
			{
				RECT rc;
				GetClientRect(label->getBox(), &rc);
				x = xParseCoord(A, 0, rc.right-rc.left);
				y = xParseCoord(B, 0, rc.bottom-rc.top);
				w = xParseDimen(C, 0, rc.right-rc.left);
				h = xParseDimen(D, 0, rc.bottom-rc.top);
			}
			int step = atoi( E );
			int time = atoi( F );

			label->reposition(3, x, y, w, h, step, time);
		}
	}
}

// !LabelRepositionBy <label> <cx> <cy> <cwidth> <cheight> [<steps>] [<time>]
void BangRepositionBy(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], A[16], B[16], C[16], D[16], E[16], F[16];
	char *buffers[] = {name, A, B, C, D, E, F};

	if (LCTokenize(pszArgs, buffers, 7, NULL) >= 5)
	{
		Label *label = lookupLabel(name);
		
		int x = atoi( A );
		int y = atoi( B );
		int w = atoi( C );
		int h = atoi( D );

		int step = atoi( E );
		int time = atoi( F );

		if (label)
			label->reposition(3, x, y, w, h, step, time, true);
		else
		{
			for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if ( stricmp((*lit)->labelGroup.c_str(), name) == NULL )
					(*lit)->reposition(3, x, y, w, h, step, time, true);
			}
		}
	}
}


// !LabelResize <label> <width> <height> [<steps> <time>]
void BangResize(HWND hwndCaller, LPCSTR pszArgs)
{
	int w = 64;
	int h = 64;
	char name[64], A[16], B[16], C[16], D[16];
	char *buffers[] = {name, A, B, C, D};

	if (LCTokenize(pszArgs, buffers, 5, NULL) >= 3)
	{
		Label *label = lookupLabel(name);

		if (label)
		{
			if (label->getBox() == NULL)
			{
				w = xParseDimen(A, 0, screenX);
				h = xParseDimen(B, 0, screenY);
			}
			else
			{
				RECT rc;
				GetClientRect(label->getBox(), &rc);
				w = xParseDimen(A, 0, rc.right-rc.left);
				h = xParseDimen(B, 0, rc.bottom-rc.top);
			}
			int step = atoi( C );
			int time = atoi( D );

			label->reposition(2, 0, 0, w, h, step, time);
		}
	}
}

// !LabelResizeBy <label> <cx> <cy> [<steps> <time>]
void BangResizeBy(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], A[16], B[16], C[16], D[16];
	char *buffers[] = {name, A, B, C, D};

	if (LCTokenize(pszArgs, buffers, 5, NULL) >= 3)
	{
		Label *label = lookupLabel(name);

		int cx = atoi( A );
		int cy = atoi( B );

		int step = atoi( C );
		int time = atoi( D );

		if (label)
			label->reposition(2, 0, 0, cx, cy, step, time, true);
		else
		{
			for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if ( stricmp((*lit)->labelGroup.c_str(), name) == NULL )
					(*lit)->reposition(2, 0, 0, cx, cy, step, time, true);
			}
		}
	}
}

// !LabelScroll <label> <'on', 'off', 'toggle', 'freeze', limit> <scrollMode(1-4)>
void BangScroll(HWND hwndCaller, LPCSTR pszArgs)
{
	// FIXME: This should be split into multiple bang commands
	char name[64], option[16], mode[16];
	char *buffers[] = {name, option, mode};

	if (LCTokenize(pszArgs, buffers, 3, NULL) >= 2)
	{
		Label *label = lookupLabel(name);

		if (label)
		{
			if (stricmp(option, "off") == NULL)
			{
				label->setScrolling(0);
				if (label->frozen)
					label->frozen = false;
			}
			else if (stricmp(option, "on") == NULL)
			{
				int scrollmode = atoi(mode);
				if (scrollmode > 0 && scrollmode < 5)
					label->setScrolling(scrollmode);
				else
					label->setScrolling(label->scrollBackup);
			}
			else if (stricmp(option, "toggle") == NULL)
			{
				if (label->getScrolling() > 0)
				{
					label->setScrolling(0);
					if (label->frozen)
						label->frozen = false;
				}
				else
				{
					int scrollmode = atoi(mode);
					if (scrollmode > 0 && scrollmode < 5)
						label->setScrolling(scrollmode);
					else
						label->setScrolling(label->scrollBackup);
				}
			}
			else if (stricmp(option, "freeze") == NULL)
			{
				if (label->frozen)
					label->frozen = false;
				else
					label->frozen = true;
			}
			else
			{
				int limit = atoi(option);
				if (limit >= 0) 
					label->setScrollLimit(limit);
			}
		}
	}
}

// !LabelSetFontColor <label> <red> <green> <blue> OR !LabelSetFontColor <label> <hexcolor>
void BangSetFontColor(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], A[16], B[16], C[16];
	char *buffers[] = {name, A, B, C};
	int numTokens = LCTokenize(pszArgs, buffers, 4, NULL);
	int color;

	if (numTokens >= 2)
	{
		Label *label = lookupLabel(name);

		if (label)
		{
			if (numTokens >= 4)
			{
				int r = atoi(A);
				int g = atoi(B);
				int b = atoi(C);

				color = RGB(r, g, b);

				string setcol = A;
				setcol.append( " " );
				setcol.append( B );
				setcol.append( " " );
				setcol.append( C );
				PrefixedLSSetVariable( name, "FontColor", setcol );
			}
			else
			{
				color = strtol(A, NULL, 16);
				color = RGB(GetBValue(color), GetGValue(color), GetRValue(color));

				PrefixedLSSetVariable( name, "FontColor", (string) A );
			}

			label->getFont()->setColor(color);
			label->repaint();
		}
		else
		{
			for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if ( stricmp((*lit)->labelGroup.c_str(), name) == NULL )
				{
					if (numTokens >= 4)
					{
						int r = atoi(A);
						int g = atoi(B);
						int b = atoi(C);

						color = RGB(r, g, b);

						string setcol = A;
						setcol.append( " " );
						setcol.append( B );
						setcol.append( " " );
						setcol.append( C );
						PrefixedLSSetVariable( name, "FontColor", setcol );
					}
					else
					{
						color = strtol(A, NULL, 16);
						color = RGB(GetBValue(color), GetGValue(color), GetRValue(color));

						PrefixedLSSetVariable( name, "FontColor", (string) A );
					}

					(*lit)->getFont()->setColor(color);
					(*lit)->repaint();
				}
			}
		}
	}
}

// !LabelSetFontShadowColor <label> <red> <green> <blue> OR !LabelSetFontColor <label> <hexcolor>
void BangSetFontShadowColor(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], A[16], B[16], C[16];
	char *buffers[] = {name, A, B, C};
	int numTokens = LCTokenize(pszArgs, buffers, 4, NULL);
	int color;

	if (numTokens == 1)
	{
		Label *label = lookupLabel(name);

		if (label)
		{
			label->getFont()->setShadow(false);
			label->repaint();

			PrefixedLSSetVariable( name, "FontShadow", "false" );
		}
		else
		{
			for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if ( stricmp((*lit)->labelGroup.c_str(), name) == NULL )
				{
					(*lit)->getFont()->setShadow(false);
					(*lit)->repaint();

					PrefixedLSSetVariable( name, "FontShadow", "false" );
				}
			}
		}
	}
	else if (numTokens >= 2)
	{
		Label *label = lookupLabel(name);

		if (label)
		{
			if (numTokens >= 4)
			{
				int r = atoi(A);
				int g = atoi(B);
				int b = atoi(C);

				color = RGB(r, g, b);

				string setcol = A;
				setcol.append( " " );
				setcol.append( B );
				setcol.append( " " );
				setcol.append( C );
				PrefixedLSSetVariable( name, "FontShadowColor", setcol );
			}
			else
			{
				color = strtol(A, NULL, 16);
				color = RGB(GetBValue(color), GetGValue(color), GetRValue(color));

				PrefixedLSSetVariable( name, "FontShadowColor", (string) A );
			}

			label->getFont()->setColor(color, true);
			label->getFont()->setShadow(true);
			label->repaint();

			PrefixedLSSetVariable( name, "FontShadow", "true" );
		}
		else
		{
			for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if( stricmp((*lit)->labelGroup.c_str(), name) == NULL )
				{
					if (numTokens >= 4)
					{
						int r = atoi(A);
						int g = atoi(B);
						int b = atoi(C);

						color = RGB(r, g, b);

						string setcol = A;
						setcol.append( " " );
						setcol.append( B );
						setcol.append( " " );
						setcol.append( C );
						PrefixedLSSetVariable( name, "FontShadowColor", setcol );
					}
					else
					{
						color = strtol(A, NULL, 16);
						color = RGB(GetBValue(color), GetGValue(color), GetRValue(color));

						PrefixedLSSetVariable( name, "FontShadowColor", (string) A );
					}

					(*lit)->getFont()->setColor(color, true);
					(*lit)->getFont()->setShadow(true);
					(*lit)->repaint();

					PrefixedLSSetVariable( name, "FontShadow", "true" );
				}
			}
		}
	}
}

// !LabelSetText <label> <text>
void BangSetText(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], editbox[MAX_LINE_LENGTH], text[MAX_LINE_LENGTH];
	char *buffers[] = {name, editbox, text};
	int numTokens = LCTokenize(pszArgs, buffers, 3, NULL);

	if (numTokens >= 1)
	{
		Label *label = lookupLabel(name);

		if ( numTokens == 3 && stricmp(editbox, ".editbox") == NULL )
		{
			if (label)
				label->setText( text, true, true);
		}
		else
		{
			if (label)
				label->setText(((numTokens >= 2) ? editbox : ""), true);
		}
	}
}

// !LabelToggle <label>
void BangToggle(HWND hwndCaller, LPCSTR pszArgs)
{
	bool textEditBox = false;
	char name[64], textbox[64];
	char *buffers[] = {name, textbox};

	LCTokenize(pszArgs, buffers, 2, NULL);

	if ( stricmp(textbox, ".editbox") == NULL )
		textEditBox = true;

	Label *label = lookupLabel(name);
	
	if (label)
	{
		if ( textEditBox )
		{
			if (label->isTextboxVisible())
				label->hide(false, true);
			else
				label->show(false, true);
		}
		else
		{
			if (label->isVisible())
				label->hide(true);
			else
				label->show(true);
		}
	}
	else
	{
		for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
		{
			if ( stricmp((*lit)->labelGroup.c_str(), name) == NULL )
			{
				if ( textEditBox )
				{
					if ((*lit)->isTextboxVisible())
						(*lit)->hide(false, true);
					else
						(*lit)->show(false, true);
				}
				else
				{
					if ((*lit)->isVisible())
						(*lit)->hide(true);
					else
						(*lit)->show(true);
				}
			}
		}
	}
}

// !LabelUpdate <label>
void BangUpdate(HWND hwndCaller, LPCSTR pszArgs)
{
	Label *label = lookupLabel(pszArgs);

	if (label)
		label->update();
	else
	{
		for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
		{
			if ( stricmp((*lit)->labelGroup.c_str(), pszArgs) == NULL )
				(*lit)->update();
		}
	}
}

// !LabelSetAlpha <label> <value>
void BangSetAlpha(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], factor[16];
	char *buffers[] = {name, factor};

	bool wasVisible = false;

	if (LCTokenize(pszArgs, buffers, 2, NULL) == 2)
	{
		Label *label = lookupLabel(name);
		
		if (label)
		{
			if ( !label->isLayeredFlagSet() )
			{
				if (label->isVisible())
				{
					wasVisible = true;
					label->hide();
				}
				else 
					wasVisible = false;
			}

			label->setAlpha(atoi( factor ));

			if (wasVisible)
				label->show();
		}
		else
		{
			for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if ( stricmp((*lit)->labelGroup.c_str(), name) == NULL )
				{
					if ( !(*lit)->isLayeredFlagSet() )
					{
						if ((*lit)->isVisible())
						{
							wasVisible = true;
							(*lit)->hide();
						}
						else 
							wasVisible = false;
					}

					(*lit)->setAlpha(atoi( factor ));

					if (wasVisible)
						(*lit)->show();
				}
			}
		}
	}
}

// !LabelGhosted <label> <on|true, off|false, toggle>
void BangGhosted(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], option[32];
	char *buffers[] = {name, option};

	LCTokenize(pszArgs, buffers, 2, NULL);

	Label *label = lookupLabel(name);

	if (label)
	{
		if (stricmp(option, "true") == NULL || stricmp(option, "on") == NULL)
			label->setGhosted(true);
		else if (stricmp(option, "false") == NULL || stricmp(option, "off") == NULL)
			label->setGhosted(false);
		else if (stricmp(option, "toggle") == NULL || stricmp(option, "switch") == NULL)
			label->setGhosted(!label->isGhosted());
	}
	else
	{
		for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
		{
			if ( stricmp((*lit)->labelGroup.c_str(), name) == NULL )
			{
				if (stricmp(option, "true") == NULL || stricmp(option, "on") == NULL)
					(*lit)->setGhosted(true);
				else if (stricmp(option, "false") == NULL || stricmp(option, "off") == NULL)
					(*lit)->setGhosted(false);
				else if (stricmp(option, "toggle") == NULL || stricmp(option, "switch") == NULL)
					(*lit)->setGhosted(!(*lit)->isGhosted());
			}
		}
	}
}

//Parse Current Evars In Bang
//!ParseEvars !anybang with evars escape code '%#' and TextEscape Sequences with %[...]%
void BangParseEvars(HWND hwndCaller, LPCSTR pszArgs)
{
	CString worker1 = pszArgs;
	CString worker2 = "";

	worker1.Replace("%#", "$");
	worker1.Replace("%|#", "$");

	while ( worker1.Find("%[") != -1 && worker1.Find("]%") != -1 )
	{
		worker2 = worker1;
		worker2 = worker2.Right(worker2.GetLength()-worker2.Find("%[")-1);
		worker2 = worker2.Left(worker2.Find("]%")+1);
		
		bool dynamic = false;
		if ( stricmp(worker2, "[labelname]") == NULL || stricmp(worker2, "[labeltext]") == NULL)
		{
			for (LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if( (*lit)->gethWnd() == hwndCaller)
				{
					worker1.Replace( ("%"+worker2+"%"), (systemInfo->processLabelText( (string)worker2, (*lit), &dynamic)).c_str() );
					break;
				}
			}
		}
		else
		{
			worker1.Replace( ("%"+worker2+"%"), (systemInfo->processLabelText( (string)worker2, NULL, &dynamic)).c_str() );
		}
	}

	xLSExecute(hwndCaller, worker1, SW_SHOWNORMAL);
}

//Set Evar
//!SetEvar "'$'evar'$'" "value"
void BangSetEvars(HWND hwndCaller, LPCSTR pszArgs)
{
	char evar[64], value[MAX_PATH_LENGTH];
	char *buffers[] = {evar};

	if (LCTokenize(pszArgs, buffers, 1, value) >= 1)
		PrefixedLSSetVariable( (string)evar, "", (string)value );
}

//Play Wav
//!PaySound "Path to wav|mp3|mid"
void BangPlaySound(HWND hwndCaller, LPCSTR pszArgs)
{
	CString m_Path = pszArgs;
	m_Path.MakeLower();

	if (m_Path.Find(".wav") || m_Path.Find(".mp3") || m_Path.Find(".mid"))
	{
		if (m_hWndSoundFile != NULL)
		{
			MCIWndStop(m_hWndSoundFile);
			MCIWndDestroy(m_hWndSoundFile);
			m_hWndSoundFile = NULL;
		}

		HANDLE hFindHandle;
		WIN32_FIND_DATA FindData;

		hFindHandle = FindFirstFile(m_Path, &FindData);

		if (INVALID_HANDLE_VALUE != hFindHandle)
		{
			m_hWndSoundFile = MCIWndCreate(NULL, NULL, WS_MINIMIZE | MCIWNDF_NOMENU, m_Path);
			
			MCIWndPlay(m_hWndSoundFile);

			SetTimer(messageHandler, NULL, 1000, 0);
		}

		FindClose(hFindHandle);
	}
	else if (m_hWndSoundFile != NULL)
	{
		MCIWndStop(m_hWndSoundFile);
		MCIWndDestroy(m_hWndSoundFile);
		m_hWndSoundFile = NULL;
	}
}